from dataclasses import dataclass, field
from typing import List
from holistics_dbt.aml_dimension import AmlDimension

@dataclass
class AmlTableModel:
  type: str = 'table'
  name: str = ''
  label: str = ''
  description: str = ''
  data_source_name: str = ''
  owner: str = ''
  table_name: str = ''
  dimensions: List[AmlDimension] = field(default_factory=list)

  def to_s(self) -> str:
    str = f"""
Model {self.name} {{
  type: 'table'
  label: '{self.label}'
  description: '{self.description}'
  owner: '{self.owner}'
  data_source_name: '{self.data_source_name}'
  table_name: '{self.table_name}'
"""

    for dimension in self.dimensions:
      str += f"""\n{dimension.to_s()}\n"""
    
    str += '}'

    return str




from enum import Enum
from dataclasses import dataclass, field

class DimensionType(Enum):
  TEXT = 'text'
  NUMBER = 'number'
  DATE = 'date'
  DATETIME = 'datetime'
  TRUEFALSE = 'truefalse'
  JSON = 'json'
  UNKNOWN = 'unknown'

class DefinitionType(Enum):
  SQL = 'sql'
  AML = 'aml'
  AQL = 'aql'

@dataclass
class DimensionDefinition:
  type: DefinitionType = None
  content: str = ''

  def to_s(self) -> str:
    return f'@{self.type} {{{{ {self.content} }}}};; '

@dataclass
class AmlDimension:
  label: str = ''
  name: str = ''
  hidden: bool = False
  description: str = ''
  type: DimensionType = None
  definition: DimensionDefinition = None
  format: str = None

  def to_s(self) -> str:
    str = f"""  dimension {self.name} {{
    type: '{self.type}'
    hidden: { 'true' if self.hidden else 'false'}
    definition: {self.definition.to_s()}"""

    if self.description:
      str += f"""
    description: '{self.description}'"""
    
    if self.label:
      str += f"""
    label: '{self.label}'"""
    
    if self.format:
      str += f"""
    format: '{self.format}'"""

    str += """
  }"""

    return str


from holistics_dbt.dbt import Dbt
from holistics_dbt.aml_table_model import AmlTableModel
from holistics_dbt.aml_dimension import AmlDimension, DimensionDefinition
from typing import List
import re
# TODO: improve this
from holistics_dbt.databases import BigQuery, BaseDatabase

def database_for(dbname: str) -> BaseDatabase:
  if dbname == 'bigquery':
    return BigQuery()
  else:
    raise ValueError(f'Unsupported database: {dbname}')

def fqname_for(schema: str, table: str, dbname: str) -> str:
  # TODO: empty schema?
  quote = database_for(dbname).quote()
  return f'{quote}{schema}{quote}.{quote}{table}{quote}'

pattern = re.compile(r'(?<!^)(?=[A-Z])')
def snake_case(s: str) -> str:
  lower_case = pattern.sub('_', s).lower()
  removed_dot = lower_case.replace('.', '_')
  return removed_dot

class AmlGen:
  def __init__(
      self,
      dbt_artifact: Dbt,
    ) -> None:
    self.dbt_artifact = dbt_artifact

  def gen_table_models(self) -> List[AmlTableModel]:
    nodes = self.dbt_artifact.catalog_nodes()
    table_models = []

    dbtype = self.dbt_artifact.adapter_type()

    for key in nodes:
      node = nodes[key]
      table_models.append(self.to_table_model(dbtype, node))

    return table_models
    
  def to_table_model(self, dbtype: str, catalog_node: dict) -> AmlTableModel:
    metadata = catalog_node['metadata']
    fqname = fqname_for(metadata['schema'], metadata['name'], dbtype)
    model_name = snake_case(catalog_node['unique_id'])

    database = database_for(dbtype)

    dimensions = map(
      lambda kv: self.to_dimension(kv[1], database),
      catalog_node['columns'].items()
    )

    return AmlTableModel(
      name=model_name,
      data_source_name=metadata['database'],
      table_name=fqname,
      description='', # TODO: search in manifest
      dimensions=dimensions,
      owner='',
    )
  
  def to_dimension(self, column: dict, database: BaseDatabase) -> AmlDimension:
    definition = DimensionDefinition(
      type='sql',
      content=f'#SOURCE.{column["name"]}'
    )

    return AmlDimension(
      label=column['name'],
      name=snake_case(column['name']),
      hidden=False, # TODO: sync with current existing file?
      description=column['comment'], # TODO: escape quotes
      definition=definition,
      format=None, # TODO: sync with current existing file?
      type=database.db_type_to_aml_type(column['type']),
    )



from abc import ABC, abstractmethod

class BaseDatabase(ABC):
  @abstractmethod
  def quote(self) -> str:
    pass
  
  @abstractmethod
  def db_type_to_aml_type(self, db_type: str) -> str:
    pass




import json

class Dbt:
  def __init__(self, manifest_path: str, catalog_path: str) -> None:
    self.manifest = dict()
    self.catalogs = dict()

    with open(manifest_path, 'r') as f:
      manifest = json.load(f)
      self.manifest['metadata'] = manifest.get('metadata')
      self.manifest['nodes'] = manifest.get('nodes')

    with open(catalog_path, 'r') as f:
      catalogs = json.load(f)
      self.catalogs['metadata'] = catalogs.get('metadata')
      self.catalogs['nodes'] = catalogs.get('nodes')

  def manifest_nodes(self) -> dict:
    return self.manifest.get('nodes')
  
  def catalog_nodes(self) -> dict:
    return self.catalogs.get('nodes')

  def adapter_type(self) -> str:
    return self.manifest['metadata']['adapter_type']



import sys
from holistics_dbt.aml_gen import AmlGen
from holistics_dbt.dbt import Dbt

dbt = Dbt(
  manifest_path='tests/manifest.json',
  catalog_path='tests/catalog.json'
)

aml_gen = AmlGen(dbt_artifact=dbt)

table_models = aml_gen.gen_table_models()

for table_model in table_models:
  print(table_model.to_s())


from holistics_dbt.databases.base_database import BaseDatabase

class BigQuery(BaseDatabase):
  def quote(self) -> str:
    return '`'
  
  def db_type_to_aml_type(self, db_type: str) -> str:
      if db_type == 'INT64' or db_type == 'FLOAT64':
        return 'number'
      elif db_type == 'BOOL':
        return 'truefalse'
      elif db_type == 'STRING':
        return 'text'
      elif db_type == 'DATE':
        return 'date'
      elif db_type == 'DATETIME':
        return 'datetime'
      else:
        return 'text'